@@ -1,10 +1,8 @@
+/* A Bison parser, made by GNU Bison 2.5.  */
 
-/* A Bison parser, made by GNU Bison 2.4.1.  */
-
-/* Skeleton interface for Bison's Yacc-like parsers in C
+/* Bison interface for Yacc-like parsers in C
    
-      Copyright (C) 1984, 1989, 1990, 2000, 2001, 2002, 2003, 2004, 2005, 2006
-   Free Software Foundation, Inc.
+      Copyright (C) 1984, 1989-1990, 2000-2011 Free Software Foundation, Inc.
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -129,11 +127,12 @@
      T_ERROR_QUEUE_LENGTH = 345,
      T_OPTIONS = 346,
      T_TCP_WINDOW_TRACKING = 347,
-     T_IP = 348,
-     T_PATH_VAL = 349,
-     T_NUMBER = 350,
-     T_SIGNED_NUMBER = 351,
-     T_STRING = 352
+     T_EXPECT_SYNC = 348,
+     T_IP = 349,
+     T_PATH_VAL = 350,
+     T_NUMBER = 351,
+     T_SIGNED_NUMBER = 352,
+     T_STRING = 353
    };
 #endif
 /* Tokens.  */
@@ -227,11 +226,12 @@
 #define T_ERROR_QUEUE_LENGTH 345
 #define T_OPTIONS 346
 #define T_TCP_WINDOW_TRACKING 347
-#define T_IP 348
-#define T_PATH_VAL 349
-#define T_NUMBER 350
-#define T_SIGNED_NUMBER 351
-#define T_STRING 352
+#define T_EXPECT_SYNC 348
+#define T_IP 349
+#define T_PATH_VAL 350
+#define T_NUMBER 351
+#define T_SIGNED_NUMBER 352
+#define T_STRING 353
 
 
 
@@ -240,7 +240,7 @@
 typedef union YYSTYPE
 {
 
-/* Line 1676 of yacc.c  */
+/* Line 2068 of yacc.c  */
 #line 53 "read_config_yy.y"
 
 	int		val;
@@ -248,7 +248,7 @@ typedef union YYSTYPE
 
 
 
-/* Line 1676 of yacc.c  */
+/* Line 2068 of yacc.c  */
 #line 253 "read_config_yy.h"
 } YYSTYPE;
 # define YYSTYPE_IS_TRIVIAL 1
